#include <stdio.h>
#include <iostream>
#include <cuda.h>

void test(int x, const char * s)
{
	int t = x;
	if (t != 0)
	{
		std::cout << "fail " << t << " " << s << "\n";
		std::cout.flush();
	}
}

#define ALIGN_UP(offset, alignment) (offset) = ((offset) + (alignment) - 1) & ~((alignment) - 1)


int main(int argc, char *argv[])
{
	argc--; argv++;

	test(cuInit(0), "cuInit");

	int deviceCount = 0;
	test(cuDeviceGetCount(&deviceCount), "cuDeviceGetCount");

	int device = 0;
	if (argc)
		device = atoi(*argv);

	CUdevice cuDevice = 0;
	test(cuDeviceGet(&cuDevice, device), "cuDeviceGet");

	CUcontext cuContext;
	int xxx = cuCtxCreate(&cuContext, 0, cuDevice);

	CUmodule cuModule;
	test(cuModuleLoad(&cuModule, "inst.ptx"), "cuModuleLoad");

	// Do basic test.  No sense continuing if we cannot complete this
	// test.
	try
	{
		CUfunction proc;
		test(cuModuleGetFunction(&proc, cuModule, "SetpFloats"), "cuModuleGetFunction");

		bool * h_B = (bool*)malloc(20 * sizeof(bool));
		memset(h_B, 0, 20 * sizeof(bool));
		CUdeviceptr d_B;
		test(cuMemAlloc(&d_B, 20 * sizeof(bool)), "cuMemAlloc");
		test(cuMemcpyHtoD(d_B, h_B, 20 * sizeof(bool)), "cuMemcpyHtoD");


		unsigned int * h_U = (unsigned int*)malloc(20 * sizeof(unsigned int));
		memset(h_U, 0, 20 * sizeof(unsigned int));
		CUdeviceptr d_U;
		test(cuMemAlloc(&d_U, 20 * sizeof(unsigned int)), "cuMemAlloc");
		test(cuMemcpyHtoD(d_U, h_U, 20 * sizeof(unsigned int)), "cuMemcpyHtoD");


		float * h_F = (float*)malloc(20 * sizeof(float));
		memset(h_F, 0, 20 * sizeof(float));
		CUdeviceptr d_F;
		test(cuMemAlloc(&d_F, 20 * sizeof(float)), "cuMemAlloc");
		test(cuMemcpyHtoD(d_F, h_F, 20 * sizeof(float)), "cuMemcpyHtoD");


		int offset = 0;
		void* ptr;
	
		ptr = (void*)(size_t)d_B;
		ALIGN_UP(offset, __alignof(ptr));
		test(cuParamSetv(proc, offset, &ptr, sizeof(ptr)), "cuParamSetv");
		offset += sizeof(ptr);

		ptr = (void*)(size_t)d_U;
		ALIGN_UP(offset, __alignof(ptr));
		test(cuParamSetv(proc, offset, &ptr, sizeof(ptr)), "cuParamSetv");
		offset += sizeof(ptr);

		ptr = (void*)(size_t)d_F;
		ALIGN_UP(offset, __alignof(ptr));
		test(cuParamSetv(proc, offset, &ptr, sizeof(ptr)), "cuParamSetv");
		offset += sizeof(ptr);

		test(cuParamSetSize(proc, offset), "cuParamSetSize");

		int threadsPerBlock = 1;
		int blocksPerGrid = 1;

		test(cuFuncSetBlockShape(proc, threadsPerBlock, 1, 1), "cuFuncSetBlockShape");

		test(cuLaunchGrid(proc, blocksPerGrid, 1), "cuLaunchGrid");

		test(cuMemcpyDtoH(h_B, d_B, 20 * sizeof(bool)), "cuMemcpyDtoH");
		test(cuMemcpyDtoH(h_U, d_U, 20 * sizeof(unsigned int)), "cuMemcpyDtoH");
		test(cuMemcpyDtoH(h_F, d_F, 20 * sizeof(float)), "cuMemcpyDtoH");

		test(cuMemFree(d_B), "cuMemFree");
		test(cuMemFree(d_U), "cuMemFree");
		test(cuMemFree(d_F), "cuMemFree");

		std::cout << "Floats " << h_F[0] << " " << h_F[1] << " " << h_F[2] << " "
			<< h_F[3] << " " << h_F[4] << "\n";
		std::cout << "uints  " << std::hex << h_U[0] << " " << h_U[1] << " " << h_U[2] << " "
			<< std::hex << h_U[3] << " " << h_U[4] << "\n";
		std::cout << "Compare using setp.ne.f32 " << h_B[0] << "\n";
		std::cout << "Compare using setp.ne.b32 " << h_B[1] << "\n";
		std::cout << "Compare using setp.ne.f32 hardwired " << h_B[2] << "\n";
		std::cout << "Compare using setp.ne.b32 hardwired " << h_B[3] << "\n";
		std::cout << "Compare using setp.ne.f32 globals " << h_B[4] << "\n";
		std::cout << "Compare using setp.ne.b32 globals " << h_B[5] << "\n";

		std::cout << "Testp.normal " << h_B[6] << "\n";
		std::cout << "Testp.notnormal " << h_B[7] << "\n";
		std::cout << "Testp.number " << h_B[8] << "\n";
		std::cout << "Testp.notanumber " << h_B[9] << "\n";

		// Now compare using CPU:
		std::cout << "Compare CPU\n";
		float one;
		float one_plus_fraction;
		typedef union {
			unsigned int U;
			float F;
		} u;
		u * p_one = (u*)&one;
		u * p_one_plus_fraction = (u*)&one_plus_fraction;
		p_one->U = 0x3f800000;
		p_one_plus_fraction->U = 0x3f800001;

		std::cout << "Floats " << p_one->F << " " << p_one_plus_fraction->F << "\n";
		std::cout << "uints  " << p_one->U << " " << p_one_plus_fraction->U << "\n";
		std::cout << "Compare floats " << (p_one->F != p_one_plus_fraction->F) << "\n";
		std::cout << "Compare uints  " << (p_one->U != p_one_plus_fraction->U) << "\n";

	} catch (...)
	{
		test(1, "test crashed.");
	}
	return 0;
}
